function [ result ] = cheb_poly_dct( freq, val, z, M, transp_flag )
%CHEB_POLY_DCT Direct evaluation of multivariate algebraic polynomial
%given in Chebyshev basis at rank-1 Chebyshev lattice nodes or computation
%of its adjoint.
%
%   This means that the evaluation nodes are then given by
%   x_j := (cos(pi*j*z(1)/M),...,cos(pi*j*z(d))/M),
%   j=0,...,M.
%
%   Input parameters:
%      freq - frequency index set as matrix of size [num_freq,d]
%      val - Chebyshev coefficients as a column vector of length num_freq
%            or function values a column vector of length M+1 (adjoint case)
%      z - generating vector of rank-1 Chebyshev lattice as a row vector of
%          length d
%      M - size parameter of rank-1 Chebyshev lattice
%      transp_flag - either 'notransp' (default) for non-adjoint
%                    or 'transp' for adjoint
%
%    Ouput parameters:
%       result - Results of function evaluations at rank-1 Chebyshev lattice
%                nodes (non-adjoint case) or coefficients (adjoint case)
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 5
  transp_flag = 'notransp';
end

nodes = cos((0:M)'*z*pi/M);

if strcmp(transp_flag, 'notransp')
  result = zeros(size(nodes,1), 1);

  for ix = 1:size(nodes,1)
    x = nodes(ix,:);
    result(ix) = sum(val .* prod(cos(freq.*repmat(acos(x),size(freq,1),1)),2));
  end
elseif strcmp(transp_flag, 'transp')
  result = zeros(size(freq,1), 1);
  for ix = 1:size(nodes,1)
    x = nodes(ix,:);
    result = result + val(ix) * prod(cos(freq.*repmat(acos(x),size(freq,1),1)),2);
  end
else
  error('unknown trans_flag: %s', transp_flag);
end

end

