function [ result ] = cheb_poly_eval( freq, val, nodes_or_zMxx )
%CHEB_POLY_EVAL Evaluates multivariate algebraic polynomial in Chebyshev basis
%given by its frequencies and corresponding Fourier coefficients.
%   Input parameter:
%      freq - matrix of frequencies with size [num_freq,d]
%      val - array of corresponding Fourier coefficents
%      x_or_zMxx - evaluation nodes
%
%   Input x_or_zMxx is either a matrix with size [num_nodes,d] or a
%   cell with three elements. In the latter case
%      x_or_zMxx{1} - generating vector z of rank-1 Chebyshev lattice of
%                     length t, 1<=t<=d
%      x_or_zMxx{2} - rank-1 Chebyshev lattice size parameter M
%      x_or_zMxx{3} - array (x_{t+1}',...,x_{d}) of length d-t
%                     containing fixed components of the evaluation vector
%      This means that the evaluation nodes are then given by
%      x_j := (cos(pi*j*z(1)/M),...,cos(pi*j*z(t))/M,x_{t+1}',...,x_{d}),
%      j=0,...,M.
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


nodes = [];
if isa(nodes_or_zMxx,'cell')
  if length(nodes_or_zMxx) == 3
    z = nodes_or_zMxx{1};
    M = nodes_or_zMxx{2};
    xx = nodes_or_zMxx{3};
    xx = xx(:);
    if isempty(xx)
      result = cheb_poly_fct(freq,val,z,M);
    else
      result = cheb_poly_fct(freq(:,1:length(z)),val .* prod(cos(freq(:,length(z)+1:size(freq,2)).*repmat(acos(xx'),size(freq,1),1)),2) ,z,M);
%      out = lfft(freq(:,1:length(z)),z,M,val.*exp(2*pi*1i*freq(:,length(z)+1:size(freq,2))*xx),'notransp');
    end
  else
    nodes = nodes_or_zMxx{1};
  end
else
  nodes = nodes_or_zMxx;
end

if ~isempty(nodes)
  result = zeros(size(nodes,1), 1);
  for ix = 1:size(nodes,1)
    x = nodes(ix,:);
    result(ix) = sum(val .* prod(cos(freq.*repmat(acos(x),size(freq,1),1)),2));
  end
end

end

