function [ result ] = cheb_poly_fct( freq, val, z, M, transp_flag )
%CHEB_POLY_FCT Fast evaluation of multivariate algebraic polynomial
%given in Chebyshev basis at rank-1 Chebyshev lattice nodes or computation
%of its adjoint via fast single one-dimensional DCT-I based on FFT.
%
%   This means that the evaluation nodes are then given by
%   x_j := (cos(pi*j*z(1)/M),...,cos(pi*j*z(d))/M),
%   j=0,...,M.
%
%   Input parameters:
%      freq - frequency index set as matrix of size [num_freq,d]
%      val - Chebyshev coefficients as a column vector of length num_freq
%            or function values a column vector of length M+1 (adjoint case)
%      z - generating vector of rank-1 Chebyshev lattice as a row vector of
%          length d
%      M - size parameter of rank-1 Chebyshev lattice
%      transp_flag - either 'notransp' (default) for non-adjoint
%                    or 'transp' for adjoint
%
%    Ouput parameters:
%       result - Results of function evaluations at rank-1 Chebyshev lattice
%                nodes (non-adjoint case) or coefficients (adjoint case)
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 5
  transp_flag = 'notransp';
end

s = 1;

d = length(z);

m_s = [1;-1];
for t=2:d
  m_s = [reshape(repmat([1,-1], size(m_s,1), 1), 2*size(m_s,1), 1) repmat(m_s, 2, 1)];
end
m_s(m_s(:,s) < 1,:) = [];

freq_tilde = [];
val_tilde = [];

if strcmp(transp_flag, 'notransp')
  for i_m = 1:size(m_s,1)
    freq_tilde = [freq_tilde; freq .* repmat(m_s(i_m,:), size(freq,1), 1)];
    val_tilde = [val_tilde; val];
  end

  k = emod(freq_tilde*z',M)+1;

  b = accumarray(k,val_tilde,[M+1,1],@sum);
  b = b / 2^(d-1);
  b(1) = 2 * b(1);
  if M > 0
    b(end) = 2 * b(end);
  end

  result = fct1(b);
elseif strcmp(transp_flag, 'transp')
  a = val;
  a(1) = 2 * a(1);
  if M > 0
    a(end) = 2 * a(end);
  end
  val = fct1(a);

  result = zeros(size(freq,1),1);
  for i_m = 1:size(m_s,1)
    result = result + val(emod((freq .* repmat(m_s(i_m,:), size(freq,1), 1)) * z',M)+1);
  end

  result = result / 2^(d-1);
else
  error('unknown trans_flag: %s', transp_flag);
end


end

