function [ val ] = cheb_poly_reco_direct( freq, z, M, fct_handle )
%CHEB_POLY_RECO_DIRECT Slow reconstruction of Chebyshev coefficients of
%multivariate algebraic polynomial in Chebyshev basis with frequencies
%supported on index set freq from samples taken at nodes x_j of
%reconstructing rank-1 Chebyshev lattice.
%
%   This means that the sampling nodes are then given by
%   x_j := (cos(pi*j*z(1)/M),...,cos(pi*j*z(d))/M),
%   j=0,...,M.
%
%   Input parameters:
%      freq - frequency index set as matrix of size [num_freq,d]
%      z - generating vector of rank-1 Chebyshev lattice as a row vector of
%          length d
%      M - size parameter of rank-1 Chebyshev lattice
%      fct_handle - function handle (block box) which accepts sampling
%                   nodes from [-1,1]^d as an array of size [num_nodes,d]
%
%    Ouput parameters:
%       val - reconstructed Chebyshev coefficients
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

nodes = cos(repmat((0:M)',1,length(z)) .* repmat(z,M+1,1)*pi/M);
f_samples = fct_handle(nodes);
clear nodes;

val = zeros(size(freq,1),1);

f_in = f_samples;
if M >= 1
  f_in(1) = f_in(1)/2;
  f_in(end) = f_in(end)/2;
end

jj = (0:M)';

for ik = 1:size(freq,1)
  val(ik) = sum(f_in .* cos(jj/M*pi*freq(ik,:)*z'));
end

%val = f_hat(emod(freq*z',M)+1);

end

