#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <complex.h>
#include <math.h>

#ifdef __linux__
  /* for getcurtime() */
  #include <sys/time.h>
#else
  /* for clock() */
  #include <time.h>
#endif

/* check for ctrl-c interrupt via utIsInterruptPending() */
#if defined (_WIN32)
  #include <windows.h>
#elif defined (__linux__)
  #include <unistd.h>
#endif

#include "mex.h"

#if __STDC_VERSION__ >= 199901L
/* C99 code */
#define C99
  #define STATIC static
  #define INLINE inline
#else
  #define STATIC
  #define INLINE
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

/* Display constants, default is verbose */
#define DISPLAY_OFF 0
#define DISPLAY_CBC 10
#define DISPLAY_VERBOSE 255

#define DISPLAY_INTERVAL 30.0

#define OMP_Z_INCR 100000

extern int utIsInterruptPending();


/* Either 1 or 2,
   1 for rebuilding index set array for each component,
   2 for one large index set array and incremental scalar product k*z */
#define GENLATTICE_DEFAULT_ALGORITHM 1

/* if enabled, check for correctness and return error to matlab if check fails */
#undef  GENLATTICE_USE_ASSERT

/* if enabled, algorithm 2 will use transposed index set for (hopefully) better cache efficiency */
#undef  GENLATTICE_FREQ_TRANSPOSE

  #define GET_BIT(bitmap, index) get_bit_64(bitmap, index)
  #define SET_BIT(bitmap, index) set_bit_64(bitmap, index)

#ifdef GENLATTICE_USE_ASSERT
  #warning using assert
  STATIC void assert_genlattice_fail(char *cond, const char *file, unsigned int line)
  {
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:assert", "assert failed in %s at line %d: %s", file, line, cond);
  }
  #define assert_genlattice(condition) {if (!(condition)) assert_genlattice_fail(__STRING(condition), __FILE__, __LINE__);}
#endif

int convert_mxArrayCM_to_int64ArrayRM(int64_t *outArray, const mxArray *inArray, const int nOutRows, const int nOutCols)
{
  int l, t;

  if (mxGetClassID(inArray) == mxDOUBLE_CLASS)
  {
    double *dfreq = mxGetPr(inArray);
    if (dfreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = llrint(dfreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxSINGLE_CLASS)
  {
    float *ffreq = (float *) mxGetData(inArray);
    if (ffreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = llrint(ffreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT32_CLASS)
  {
    int32_T *ifreq = (int32_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT16_CLASS)
  {
    int16_T *ifreq = (int16_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT8_CLASS)
  {
    int8_T *ifreq = (int8_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }

  return 1;
}

int convert_mxArrayCM_to_intArrayRM(int *outArray, const mxArray *inArray, const int nOutRows, const int nOutCols)
{
  int l, t;

  if (mxGetClassID(inArray) == mxDOUBLE_CLASS)
  {
    double *dfreq = mxGetPr(inArray);
    if (dfreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = lrint(dfreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxSINGLE_CLASS)
  {
    float *ffreq = (float *) mxGetData(inArray);
    if (ffreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = lrint(ffreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT32_CLASS)
  {
    int32_T *ifreq = (int32_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT16_CLASS)
  {
    int16_T *ifreq = (int16_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT8_CLASS)
  {
    int8_T *ifreq = (int8_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }

  return 1;
}

/**
  * Returns current time in seconds for time measurements.
  * On Linux, the current system time (gettimeofday()) is used,
  * otherwise the processor time used by the current program (clock()).
  */
STATIC INLINE double getcurtime(void)
{
#ifdef __linux__
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return 1.0 * tv.tv_sec + 1e-6 * tv.tv_usec;
#else
  return (double) clock() / CLOCKS_PER_SEC;
#endif
}

/**
  * Calculates rem = a mod b, where 0 <= rem < b.
  */
STATIC INLINE uint64_t modl(int64_t a, int64_t b)
{
/*  lldiv_t result = lldiv(a, b);
  int64_t rem = result.rem; */
  int64_t rem = a % b;
  if (rem < 0)
    rem += b;
#ifdef GENLATTICE_USE_ASSERT
  assert_genlattice(rem >= 0 && rem < b);
#endif

  return (uint64_t) rem;
}

/**
  * Calculates rem = a mod 2^s, where 0 <= rem < 2^s.
  */
STATIC INLINE uint64_t modl2(int64_t a, uint8_t s)
{
  const uint64_t b = 1UL << s;
  uint64_t rem;

  rem = a & (b-1);
  if (rem < 0)
    rem += b;

#ifdef GENLATTICE_USE_ASSERT
    assert_genlattice(rem == modl(a, b));
#endif

  return rem;
}

/**
  * Calculates rem = a emod b, where 0 <= rem < b.
  */
STATIC INLINE uint64_t emodl(int64_t a, int64_t b)
{
  int64_t rem = a % (2L * b);
  if (rem < 0)
    rem += 2L * b;
  if (rem > b)
    rem = 2L * b - rem;
#ifdef GENLATTICE_USE_ASSERT
  if (rem < 0 || rem > b)
    printf("a=%ld, b=%ld, rem=%ld\n", a, b, rem);
  assert_genlattice(rem >= 0 && rem <= b);
#endif

  return (uint64_t) rem;
}

/** Variable used by comparator function for index set I */
STATIC int compsize = 2;

/**
  * Comparator function for sorting frequency index set.
  */
STATIC int compfunc(const void *p1, const void *p2)
{
  int t;
  const int *i1, *i2;

  i1 = (int *) p1;
  i2 = (int *) p2;

  for (t = 0; t < compsize; t++)
  {
    if (i1[t] < i2[t])
      return -1;
    else if (i1[t] > i2[t])
      return 1;
  }

  return 0;
}

/** Variable used by comparator function for index set I_tilde */
STATIC int compsize_tilde = 2;

/**
  * Comparator function for sorting frequency index set.
  */
STATIC int compfunc_tilde(const void *p1, const void *p2)
{
  int t;
  const int64_t *i1, *i2;

  i1 = (int64_t *) p1;
  i2 = (int64_t *) p2;

  if (i1[compsize_tilde] < i2[compsize_tilde])
    return -1;
  else if (i1[compsize_tilde] > i2[compsize_tilde])
    return 1;

  for (t = 0; t < compsize_tilde; t++)
  {
    if (i1[t] > 0 && i2[t] < 0)
      return -1;
    else if (i1[t] < 0 && i2[t] > 0)
      return 1;
  }

  for (t = 0; t < compsize_tilde; t++)
  {
    if (i1[t] < i2[t])
      return -1;
    else if (i1[t] > i2[t])
      return 1;
  }

  return 0;
}

/**
  * Gets bit at certain index from bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE uint64_t get_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  return temp[tindex] & tbit;
}

/**
  * Sets bit at certain index in bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE void set_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  temp[tindex] |= tbit;
}

STATIC INLINE int64_t min_i(int64_t a, int64_t b)
{
  return a < b ? a : b;
}

STATIC INLINE int64_t max_i(int64_t a, int64_t b)
{
  return a >= b ? a : b;
}

STATIC INLINE int32_t min_i32(int32_t a, int32_t b)
{
  return a < b ? a : b;
}

STATIC void flushMatlab(void)
{
  fflush(NULL);
/*  mexEvalString("drawnow;"); */
}

/* heapsort from rosettacode.org, modified for key/val */
STATIC void heapsort_siftdown( int64_t *key, int64_t *val, int64_t start, int64_t end)
{
  int64_t root = start;
 
  while (root*2+1 < end)
  {
    int child = 2 * root + 1;
    if (child + 1 < end && key[child] < key[child+1])
      child++;

    if (key[root] < key[child])
    {
      int64_t temp;
      temp = key[child]; key[child] = key[root]; key[root] = temp;
      temp = val[child]; val[child] = val[root]; val[root] = temp;
      root = child;
    }
    else
      return;
  }
}

/* heapsort from rosettacode.org, modified for key/val */
STATIC void heapsort( int64_t *key, int64_t *val, int64_t count)
{
    int64_t start, end;
 
    /* heapify */
    for (start = (count-2)/2; start >=0; start--)
    {
      heapsort_siftdown( key, val, start, count);
    }
 
    for (end = count-1; end > 0; end--)
    {
      int64_t temp;
      temp = key[end]; key[end] = key[0]; key[0] = temp;
      temp = val[end]; val[end] = val[0]; val[0] = temp;
      heapsort_siftdown(key, val, 0, end);
    }
}
 

/**
  * Searches for non-aliasing generating vector z of rank-1 lattice for given
  * frequency index set and lattice size.
  */
STATIC int search_gen_z_r1lattice(int d, int64_t *z, int *freq, int len_freq, int64_t M, int inc, int nthreads, int display)
{
  int t, tt, k, l;
  int64_t curz;

  const uint64_t bitmap_length = 1+M/(1U<<6)+1;
  const uint64_t bitmap_size = bitmap_length * sizeof(uint64_t);
  int *tempfreq = (int *) mxMalloc(d*len_freq*sizeof(int));
  double time0, time1, time2;
  int64_t *tempfreqtilde = NULL;
  uint64_t len_tempfreqtilde = 0;

  volatile bool interrupted = false;

  time0 = getcurtime();

  for (k = 0; k < len_freq; k++)
  {
    int cur_factor = 1;
    for (t = 1; t < d; t++)
    {
      if (freq[d*k+t] != 0)
        cur_factor *= 2;
    }
    len_tempfreqtilde += cur_factor;
  }

  tempfreqtilde = (int64_t *) mxMalloc(d*len_tempfreqtilde*sizeof(int64_t));

  for (t = 1; t < d; t++)
  {
    volatile bool found = false;
    uint64_t zstart = (inc == 1) ? max_i(z[t-1]+1, z[t]) : max_i(0, z[t]);
    uint64_t zInnerLoopStart;
    int64_t m_vals[d*(1U<<(d-1))];
    int itilde, im;

    time1 = getcurtime();
    compsize = t+1;
    qsort(freq, len_freq, d*sizeof(int), compfunc);

    for (tt = 0; tt <= t; tt++)
      tempfreq[tt] = freq[tt];

    for (l = 0; l < 1U<<(t); l++)
    {
      int ll = l;
      m_vals[d*l+0] = 1;
      for (tt = t; tt >= 1; tt--)
      {
        if (ll % 2 == 0)
          m_vals[d*l+tt] = 1;
	else
	  m_vals[d*l+tt] = -1;
	ll = ll/2;
      }
    }

    itilde = 0;
    for (im = 0; im < 1U<<(t); im++)
    {
#ifdef GENLATTICE_USE_ASSERT
      assert_genlattice(itilde < len_tempfreqtilde);
#endif
      for (tt = 1; tt <= t; tt++)
        if (tempfreq[tt] == 0 && m_vals[d*im+tt] == -1)
	  break;

      if (tt <= t)
        continue;

      tempfreqtilde[d*itilde+0] = tempfreq[0] * m_vals[d*im+0];
      for (tt = 1; tt <= t; tt++)
        tempfreqtilde[d*itilde+tt] = tempfreq[tt] * m_vals[d*im+tt];

      itilde++;
    }

    l = 1;
    for (k = 1; k < len_freq; k++)
    {
      int cc = 0;
      for (tt = 0; tt <= t; tt++)
        if (tempfreq[d*(l-1)+tt] == freq[d*k+tt])
	  cc++;

      if (cc == tt)
        continue;

      for (tt = 0; tt <= t; tt++)
        tempfreq[d*l+tt] = freq[d*k+tt];
      
      for (im = 0; im < 1U<<(t); im++)
      {
        for (tt = 1; tt <= t; tt++)
          if (tempfreq[d*l+tt] == 0 && m_vals[d*im+tt] == -1)
	    break;

        if (tt <= t)
          continue;

#ifdef GENLATTICE_USE_ASSERT
        assert_genlattice(itilde < len_tempfreqtilde);
#endif

        tempfreqtilde[d*itilde+0] = tempfreq[d*l+0] * m_vals[d*im+0];
        for (tt = 1; tt <= t; tt++)
          tempfreqtilde[d*itilde+tt] = tempfreq[d*l+tt] * m_vals[d*im+tt];

        itilde++;
      }

      l++;
    }

for (zInnerLoopStart = zstart; zInnerLoopStart < M+1 && !found && !interrupted; zInnerLoopStart += OMP_Z_INCR)
{
  uint64_t zInnerLoopUpper = min_i(M+1, zInnerLoopStart + OMP_Z_INCR);
  #pragma omp parallel shared(found,interrupted) private(tt,k) num_threads(nthreads)
  {
    uint64_t *bitmap;
    int64_t *tempsp, *tempsptilde, *index_tempsp, *index_tempsptilde;
    #pragma omp critical
    {
      bitmap = (uint64_t *) mxMalloc(bitmap_size);
      tempsp = (int64_t *) mxMalloc(l*sizeof(int64_t));
      index_tempsp = (int64_t *) mxMalloc(l*sizeof(int64_t));
      tempsptilde = (int64_t *) mxMalloc(itilde*sizeof(int64_t));
      index_tempsptilde = (int64_t *) mxMalloc(itilde*sizeof(int64_t));
    }

/*    for (z[t] = zstart; z[t] < M; z[t]++) */
/*    for (curz = zstart; curz < M; curz++) */
    #pragma omp for schedule(dynamic,1)
    for (curz = zInnerLoopStart; curz < zInnerLoopUpper; curz++)
    {
      int h;
      int collision = 0;

      if (found || interrupted)
        continue;

 #ifdef _OPENMP
     if (omp_get_thread_num() == 0)
 #endif
     {
      if (utIsInterruptPending() != 0) {
	mexWarnMsgTxt("CTRL-C detected. Aborting...");
        flushMatlab();

        interrupted = true;
        continue;
      }

      if (display >= DISPLAY_VERBOSE)
      {
        #pragma omp critical
	{
          time2 = getcurtime();
          if (time2 - time1 > DISPLAY_INTERVAL)
          {
            time1 = time2;
            printf("%d %ld: ", d, M);
            for (tt = 0; tt <= t; tt++)
              printf(" %ld ", z[tt]);
            printf(" ?%ld?\n", curz);
            flushMatlab();
          }
	}
      }
     }

      memset(bitmap, 0, bitmap_size);

      for (k = 0; k < l; k++)
      {
        int64_t ii = 0;
	uint64_t index;
        for (tt = 0; tt <= t-1; tt++)
	  ii += tempfreq[d*k+tt] * z[tt];
	ii += tempfreq[d*k+t] * curz;

        index = emodl(ii, M);
       
        if (GET_BIT(bitmap, index) != 0)
	  break;
	SET_BIT(bitmap, index);

        tempsp[k] = index;
        index_tempsp[k] = k;
      }

      if (k < l)
        continue;

      for (k = 0; k < itilde; k++)
      {
        int64_t ii = 0;
	uint64_t index;
        for (tt = 0; tt <= t-1; tt++)
	  ii += tempfreqtilde[d*k+tt] * z[tt];
	ii += tempfreqtilde[d*k+t] * curz;

        index = emodl(ii, M);

        tempsptilde[k] = index;
        index_tempsptilde[k] = k;
      }

      heapsort(tempsp, index_tempsp, l);
      heapsort(tempsptilde, index_tempsptilde, itilde);

      h = 0;
      for (k = 0; k < l; k++)
      {
        int index_k = index_tempsp[k];
	tt = 0;

        while (h < itilde && tempsp[k] > tempsptilde[h])
	  h++;

        while (h < itilde && tempsp[k] == tempsptilde[h])
	{
	  int index_h = index_tempsptilde[h];

	  for (tt = 0; tt <= t; tt++)
	    if (tempfreq[d*index_k+tt] != abs(tempfreqtilde[d*index_h+tt]))
	      break;

	  if (tt <= t)
	    break;

	  h++;
	}

        if (tt <= t)
	{
	  collision = 1;
	  break;
	}
      }

      if (k == l && collision == 0)
      {
        #pragma omp critical
	{
	  if(display >= DISPLAY_VERBOSE)
#ifdef _OPENMP
            printf("thread %d found z[%d]=%ld\n", omp_get_thread_num(), t, curz);
#else
            printf("found z[%d]=%ld\n", t, curz);
#endif

	  if (found)
	  {
	    if (curz < z[t])
	      z[t] = curz;
	  }
	  else
	    z[t] = curz;

          found = true;
	}
      }

      if (found || interrupted)
        continue;
    } /* end for curz */
    #pragma omp critical
    {
      mxFree(bitmap);
      mxFree(tempsp);
      mxFree(tempsptilde);
      mxFree(index_tempsp);
      mxFree(index_tempsptilde);
    }
 }
}

    if (interrupted)
    {
      mxFree(tempfreq);
      mxFree(tempfreqtilde);
      return -1;
    }

    if (!found)
    {
      mxFree(tempfreq);
      mxFree(tempfreqtilde);
      if (display >= DISPLAY_VERBOSE)
      {
        printf("Not found. Time: %.3e\n", getcurtime() - time0);
      }
      return 1;
    }

    if (display >= DISPLAY_CBC)
    {
      printf("z(%d) = %ld\n", t+1, z[t]);
      printf("z(1:%d) = [", t+1);
      for (tt = 0; tt <= t; tt++)
        printf(" %ld ", z[tt]);
      printf("]\n");
      flushMatlab();
    }
  }

  mxFree(tempfreq);
  mxFree(tempfreqtilde);

  if (display >= DISPLAY_CBC)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }
  return 0;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  int nfreq;
  int d;
  int64_t M;
  double *outZ;
  int l, t, v;
  int *freq;
  int64_t *z;
  int inc = 0;
  int algorithmNr = GENLATTICE_DEFAULT_ALGORITHM;
  int display = DISPLAY_VERBOSE;
  int *cuboid_min, *cuboid_max;
  uint64_t cuboid_size;
#ifdef _OPENMP
  int nthreads = omp_get_max_threads();
#else
  int nthreads = 1;
#endif

  if(nrhs < 2)
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:nrhs",
                      "Two inputs required. Usage: z = dct_cbc_lattice_search_mex64_openmp(frequency_index_set, lattice_size_M)");

  if(nlhs!=1)
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:nlhs",
                      "One output required. Usage: z = dct_cbc_lattice_search_mex64_openmp(frequency_index_set, lattice_size_M)");

  if(mxGetNumberOfElements(prhs[1])!= 1 )
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:M_not_scalar",
                      "Input lattice size M (second argument) must be a scalar.");

  nfreq = mxGetM(prhs[0]);
  d = mxGetN(prhs[0]);
  if (mxGetScalar(prhs[1]) > (double) (1UL<<53))
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^53.");
    
  M = llrint(mxGetScalar(prhs[1]));

  if ((nrhs > 2 && mxIsStruct(prhs[2])) || (nrhs > 3 && mxIsStruct(prhs[3])))
  {
    int index = mxIsStruct(prhs[2]) ? 2 : 3;
    char str[256];
    mxArray *field;

    field = mxGetField(prhs[index], 0, "algorithm");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Algorithm");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "algorithm2") == 0 || strcmp(str, "Algorithm2") == 0 || strcmp(str, "2") == 0)
        algorithmNr = 2;
      else if (strcmp(str, "algorithm1") == 0 || strcmp(str, "Algorithm1") == 0 || strcmp(str, "1") == 0)
        algorithmNr = 1;
    }

    field = mxGetField(prhs[index], 0, "display");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Display");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "off") == 0 || strcmp(str, "Off") == 0
        ||strcmp(str, "none") == 0 || strcmp(str, "None") == 0)
        display = DISPLAY_OFF;
      else if (strcmp(str, "cbc") == 0 || strcmp(str, "Cbc") == 0 || strcmp(str, "CBC") == 0)
        display = DISPLAY_CBC;
    }

    field = mxGetField(prhs[index], 0, "inc");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Inc");
    if (field != NULL && lrint(mxGetScalar(field)) == 0)
      inc = 0;
    else if (field != NULL && lrint(mxGetScalar(field)) == 1)
      inc = 1;

    field = mxGetField(prhs[index], 0, "nthreads");
    if (field != NULL)
    {
      nthreads = min_i32(lrint(mxGetScalar(field)), nthreads);

      if (nthreads < 1)
        nthreads = 1;
    }
  }

  if (display > DISPLAY_OFF)
  {
    printf("d = %d, card_freq_index_set = %d, lattice_size_M = %ld, inc = %d, nthreads = %d, display = ", d, nfreq, M, inc, nthreads);
    if (display == DISPLAY_CBC)
      printf("cbc");
    else if (display == DISPLAY_VERBOSE)
      printf("verbose");
    printf("\n");
    flushMatlab();
  }

  if (M < nfreq)
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:M_less_than_card",
                      "Input lattice_size_M is smaller than cardinality of frequency index set.");

  z = (int64_t *) mxMalloc(d*sizeof(int64_t));
  z[0] = 1;
  for (t = 1; t < d; t++)
    z[t] = 0;

  if ((nrhs > 2 && !mxIsStruct(prhs[2])) || (nrhs > 3 && !mxIsStruct(prhs[3])))
  {
    int index = !mxIsStruct(prhs[2]) ? 2 : 3;
/*
    double *zin = mxGetPr(prhs[index]);
    int lzin = mxGetN(prhs[index]);
    for (t = 0; t < d && t < lzin; t++)
      z[t] = lrint(zin[t]);
*/
    int lzin = mxGetN(prhs[index]);
    v = convert_mxArrayCM_to_int64ArrayRM(z, prhs[index], 1, min_i(d,lzin));
    if (v != 0)
    {
      if (v == 1)
        mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:z_unsupported_type",
                          "Input z is of unsupported type. Either use double, single, int32, int16 or int8 array.");
      else if (v == 2)
        mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:z_not_real",
                          "Input z does not consist of real values.");
      else
        mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:z_unknown_error",
                          "Input z: unknown error.");
    }
  }

  if (display > DISPLAY_OFF)
  {
    printf("Initial z: ");
    for (t = 0; t < d; t++)
      printf("%ld ", z[t]);
    printf("\n");
  }

  freq = (int *) mxMalloc(d*nfreq*sizeof(int));
/*
  if (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS)
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:freq_not_double",
                      "Input frequency_index_set not of type double.");

  dfreq = mxGetPr(prhs[0]);
  if (dfreq == NULL)
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:freq_not_real",
                      "Input frequency_index_set does not consist of real values.");

  for (l = 0; l < nfreq; l++)
    for (t = 0; t < d; t++)
      freq[d*l+t] = lrint(dfreq[nfreq*t+l]);
*/

  v = convert_mxArrayCM_to_intArrayRM(freq, prhs[0], nfreq, d);
  if (v != 0)
  {
    if (v == 1)
      mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:freq_unsupported_type",
                        "Input frequency_index_set is of unsupported type. Either use double, single, int32, int16 or int8 array.");
    else if (v == 2)
      mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:freq_not_real",
                        "Input frequency_index_set does not consist of real values.");
    else
      mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:freq_unknown_error",
                        "Input frequency_index_set: unknown error.");

  }

  cuboid_min = (int *) mxMalloc(d*sizeof(int));
  cuboid_max = (int *) mxMalloc(d*sizeof(int));
  for (t = 0; t < d; t++)
  {
    cuboid_min[t] = freq[t];
    cuboid_max[t] = freq[t];
  }
  for (l = 1; l < nfreq; l++)
    for (t = 0; t < d; t++)
    {
      cuboid_min[t] = min_i(cuboid_min[t], freq[d*l+t]);
      cuboid_max[t] = max_i(cuboid_max[t], freq[d*l+t]);
    }

  cuboid_size = 2 * (cuboid_max[0] - cuboid_min[0] + 1);
  for (t = 1; t < d; t++)
  {
    uint64_t factor = 2 * (cuboid_max[t] - cuboid_min[t] + 1);
    if (factor * cuboid_size < cuboid_size)
    {
      cuboid_size = 0;
      break;
    }
    cuboid_size *= factor;
  }

  if (cuboid_size != 0 && M > cuboid_size)
  {
    if (display > DISPLAY_OFF)
      printf("Cardinality of full cuboid-shaped grid in frequency domain: %ld, M=%ld\n", cuboid_size, M);
    mexWarnMsgTxt("Size of rank-1 lattice greater than full cuboid-shaped grid in frequency domain");
  }

    if (display > DISPLAY_OFF)
    {
      printf("Algorithm 1: Re-sort index set for each additional component, non-incremental SP, ");
      printf("64-bit map");
      printf("\n");
      flushMatlab();
    }
    v = search_gen_z_r1lattice(d, z, freq, nfreq, M, inc, nthreads, display);

  if (v == 0)
  {
    plhs[0] = mxCreateDoubleMatrix(1,d,mxREAL);
    outZ = mxGetPr(plhs[0]);
    for (t = 0; t < d; t++)
      outZ[t] = (double) z[t];

    if (display > DISPLAY_OFF)
    {
      printf("z = [", t+1);
      for (t = 0; t < d; t++)
        printf(" %ld ", z[t]);
      printf("]\n");
      flushMatlab();
    }
  }
  else if (v == 1)
  {
    plhs[0] = mxCreateDoubleMatrix(0,0,mxREAL);
  }

  mxFree(cuboid_min);
  mxFree(cuboid_max);
  mxFree(freq);
  mxFree(z);

  if (v == -1)
  {
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:ctrl_c",
                      "CTRL-C pressed. Aborted.");
  }

  if (v != 0 && v != 1)
  {
    mexErrMsgIdAndTxt("dct_cbc_lattice_search_mex64_openmp:unknown",
                      "unknown error");
  }

}

