function [ isreco ] = dct_r1l_check( freq, z, M, s )
%DCT_R1L_CHECK Checks if given rank-1 Chebyshev lattice fulfills
%reconstructing property for frequency index set freq.
%
%   Input parameters:
%      freq - frequency index set as matrix of size [num_freq,d]
%      z - generating vector of rank-1 Chebyshev lattice as a row vector of
%          length d
%      M - size parameter of rank-1 Chebyshev lattice
%      s - optional index in {1,...,d} where freq is not mirror (default 1)
%
%    Ouput parameters:
%      isreco - 1 if reconstructing property is fulfilled, 0 otherwise
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 4
  s = 1;
end

d = length(z);

isreco = 0;

%index set I
freq = unique(freq,'rows');
lfreq = size(freq,1);

hz = emod(freq*z',M);
if length(unique(hz)) ~= lfreq
  return
end

%matrix with index set I and values k*z for k\in I,
%sorted by k*z ascending and frequencies k\in I descending
freq_val = sortrows([freq, hz], [d+1, -1:-1:-d]);

m_s = [1;-1];
for t=2:d
  m_s = [reshape(repmat([1,-1], size(m_s,1), 1), 2*size(m_s,1), 1) repmat(m_s, 2, 1)];
end
m_s(m_s(:,s) < 1,:) = [];

%index set \tilde{I} = \mathcal{M}_s(I)
freqtilde = zeros(lfreq * size(m_s,1),d);
for im = 1:size(m_s,1)
  freqtilde((im-1)*lfreq+1 : im*lfreq, :) = freq .* repmat(m_s(im,:), lfreq, 1);
end
freqtilde = unique(freqtilde,'rows');

%matrix with index set \tilde{I} and values h*z for h\in\tilde{I},
%sorted by h*z ascending and frequencies h\in\tilde{I} descending
freqtilde_val = sortrows([freqtilde, emod(freqtilde*z',M), min(freqtilde,[],2)], [d+1, -d-2, -1:-1:-d]);
freqtilde_val = freqtilde_val(:,1:d+1);

%start with first row in matrix with index set \tilde{I}
ih = 1;
%for each row in matrix with index set I (for each k\in I)
for ik = 1:lfreq
  %frequency k\in I
  k = freq_val(ik,1:d);
  %k*z emod M
  k_val = freq_val(ik,d+1);

  %Search for next h\in\tilde{I} such that k*z emod M = h*z emod M.
  while k_val ~= freqtilde_val(ih,d+1)
    ih = ih + 1;
  end
  
  %Since I\cap\tilde{I} = I and due to the sorting,
  %h=freqtilde_val(ih,1:d) should be identical to k
  %(and also k*z emod M = h*z emod M).
  if freq_val(ik,:) ~= freqtilde_val(ih,:)
    error('internal error');
  end

  %Continue with frequencies h\in\tilde{I} where k*z emod M = h*z emod M
  while ((ih <= size(freqtilde,1)) && (k_val == freqtilde_val(ih,d+1)))
    %If there exists a frequency h\in\tilde{I} such that
    %k*z emod M = h*z emod M and k ~= abs(h),
    %then the reconstruction property is violated and isreco=0 is returned.
%   if ~isempty(find((k - abs(freqtilde_val(ih,1:d))) ~= 0, 1))
    if max(k ~= abs(freqtilde_val(ih,1:d)))
      return;
    end
    ih = ih + 1;
  end
end

isreco = 1;

end

