/*
 * Copyright (c) 2014 Lutz Kaemmerer, 2015 Toni Volkmer
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <complex.h>
#include <math.h>

#ifdef __linux__
  /* for getcurtime() */
  #include <sys/time.h>
#else
  /* for clock() */
  #include <time.h>
#endif

/* check for ctrl-c interrupt via utIsInterruptPending() */
#if defined (_WIN32)
  #include <windows.h>
#elif defined (__linux__)
  #include <unistd.h>
#endif

#include "mex.h"

#if __STDC_VERSION__ >= 199901L
/* C99 code */
#define C99
  #define STATIC static
  #define INLINE inline
#else
  #define STATIC
  #define INLINE
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

/* Display constants, default is verbose */
#define DISPLAY_OFF 0
#define DISPLAY_CBC 10
#define DISPLAY_VERBOSE 255

#define DISPLAY_INTERVAL 30.0

#define OMP_Z_INCR 100000

extern int utIsInterruptPending();


/** Either 1 or 2,
  * 1 for rebuilding index set array for each component,
  * 2 for one large index set array and incremental scalar product k*z
  */
#define GENLATTICE_DEFAULT_ALGORITHM 2

/* if enabled, check for correctness and return error to matlab if check fails */
#undef  GENLATTICE_USE_ASSERT

/* if enabled, algorithm 2 will use transposed index set for (hopefully) better cache efficiency */
#undef  GENLATTICE_FREQ_TRANSPOSE

  #define GET_BIT(bitmap, index) get_bit_64(bitmap, index)
  #define SET_BIT(bitmap, index) set_bit_64(bitmap, index)

#ifdef GENLATTICE_USE_ASSERT
  #warning using assert
  static void assert_genlattice_fail(char *cond, const char *file, unsigned int line)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:assert", "assert failed in %s at line %d: %s", file, line, cond);
  }
  #define assert_genlattice(condition) {if (!(condition)) assert_genlattice_fail(__STRING(condition), __FILE__, __LINE__);}
#endif

int convert_mxArrayCM_to_int64ArrayRM(int64_t *outArray, const mxArray *inArray, const int nOutRows, const int nOutCols)
{
  int l, t;

  if (mxGetClassID(inArray) == mxDOUBLE_CLASS)
  {
    double *dfreq = mxGetPr(inArray);
    if (dfreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = llrint(dfreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxSINGLE_CLASS)
  {
    float *ffreq = (float *) mxGetData(inArray);
    if (ffreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = llrint(ffreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT32_CLASS)
  {
    int32_T *ifreq = (int32_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT16_CLASS)
  {
    int16_T *ifreq = (int16_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT8_CLASS)
  {
    int8_T *ifreq = (int8_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }

  return 1;
}

int convert_mxArrayCM_to_intArrayRM(int *outArray, const mxArray *inArray, const int nOutRows, const int nOutCols)
{
  int l, t;

  if (mxGetClassID(inArray) == mxDOUBLE_CLASS)
  {
    double *dfreq = mxGetPr(inArray);
    if (dfreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = lrint(dfreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxSINGLE_CLASS)
  {
    float *ffreq = (float *) mxGetData(inArray);
    if (ffreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = lrint(ffreq[nOutRows*t+l]);

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT32_CLASS)
  {
    int32_T *ifreq = (int32_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT16_CLASS)
  {
    int16_T *ifreq = (int16_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }
  else if (mxGetClassID(inArray) == mxINT8_CLASS)
  {
    int8_T *ifreq = (int8_T *) mxGetData(inArray);
    if (ifreq == NULL)
      return 2;

    for (l = 0; l < nOutRows; l++)
      for (t = 0; t < nOutCols; t++)
        outArray[nOutCols*l+t] = ifreq[nOutRows*t+l];

    return 0;
  }

  return 1;
}

/**
  * Returns current time in seconds for time measurements.
  * On Linux, the current system time (gettimeofday()) is used,
  * otherwise the processor time used by the current program (clock()).
  */
STATIC INLINE double getcurtime(void)
{
#ifdef __linux__
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return 1.0 * tv.tv_sec + 1e-6 * tv.tv_usec;
#else
  return (double) clock() / CLOCKS_PER_SEC;
#endif
}

/**
  * Calculates rem = a mod b, where 0 <= rem < b.
  */
STATIC INLINE uint64_t modl(int64_t a, int64_t b)
{
  int64_t rem = a % b;
  if (rem < 0)
    rem += b;
#ifdef GENLATTICE_USE_ASSERT
  assert_genlattice(rem >= 0 && rem < b);
#endif

  return (uint64_t) rem;
}

/**
  * Calculates rem = a mod 2^s, where 0 <= rem < 2^s.
  */
STATIC INLINE uint64_t modl2(int64_t a, uint8_t s)
{
  const uint64_t b = 1UL << s;
  uint64_t rem;

  rem = a & (b-1);
  if (rem < 0)
    rem += b;

#ifdef GENLATTICE_USE_ASSERT
    assert_genlattice(rem == modl(a, b));
#endif

  return rem;
}

/** Variable used by comparator function */
static int compsize = 2;

/**
  * Comparator function for sorting frequency index set.
  */
static int compfunc(const void *p1, const void *p2)
{
  int t;
  const int *i1, *i2;

  i1 = (int *) p1;
  i2 = (int *) p2;

  for (t = 0; t < compsize; t++)
  {
    if (i1[t] < i2[t])
      return -1;
    else if (i1[t] > i2[t])
      return 1;
  }

  return 0;
}

/**
  * Gets bit at certain index from bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE uint64_t get_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  return temp[tindex] & tbit;
}

/**
  * Sets bit at certain index in bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE void set_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  temp[tindex] |= tbit;
}

STATIC INLINE int64_t min_i(int64_t a, int64_t b)
{
  return a < b ? a : b;
}

STATIC INLINE int64_t max_i(int64_t a, int64_t b)
{
  return a >= b ? a : b;
}

STATIC INLINE int32_t min_i32(int32_t a, int32_t b)
{
  return a < b ? a : b;
}

static void flushMatlab(void)
{
  fflush(NULL);
/*  mexEvalString("drawnow;"); */
}

/**
  * Searches for non-aliasing generating vector z of rank-1 lattice for given
  * frequency index set and lattice size.
  */
static int search_gen_z_r1lattice(int d, int64_t *z, int *freq, int len_freq, int64_t M, int inc, int nthreads, int display)
{
  int t, tt, k, l;
  int64_t curz;

  const uint64_t bitmap_length = 1+M/(1U<<6);
  const uint64_t bitmap_size = bitmap_length * sizeof(uint64_t);
  int *tempfreq = (int *) mxMalloc(d*len_freq*sizeof(int));
  double time0, time1, time2;

  int64_t M_log2 = 0;
  bool M_powerof_2 = false;

  volatile bool interrupted = false;

  if (display >= DISPLAY_VERBOSE && M_powerof_2)
  {
    printf("M = 2^%ld\n", M_log2);
    flushMatlab();
  }

  time0 = getcurtime();

  for (t = 1; t < d; t++)
  {
    volatile bool found = false;
    uint64_t zstart = (inc == 1) ? max_i(z[t-1]+1, z[t]) : max_i(0, z[t]);
    uint64_t zInnerLoopStart;

    time1 = getcurtime();
    compsize = t+1;
    qsort(freq, len_freq, d*sizeof(int), compfunc);

    for (tt = 0; tt <= t; tt++)
      tempfreq[tt] = freq[tt];
    l = 1;

    for (k = 1; k < len_freq; k++)
    {
      int cc = 0;
      for (tt = 0; tt <= t; tt++)
        if (tempfreq[d*(l-1)+tt] == freq[d*k+tt])
	  cc++;

      if (cc == tt)
        continue;

      for (tt = 0; tt <= t; tt++)
        tempfreq[d*l+tt] = freq[d*k+tt];
      
      l++;
    }

for (zInnerLoopStart = zstart; zInnerLoopStart < M && !found && !interrupted; zInnerLoopStart += OMP_Z_INCR)
{
  uint64_t zInnerLoopUpper = min_i(M, zInnerLoopStart + OMP_Z_INCR);
  #pragma omp parallel shared(found,interrupted) private(tt,k) num_threads(nthreads)
  {
    uint64_t *bitmap;
    #pragma omp critical
    {
      bitmap = (uint64_t *) mxMalloc(bitmap_size);
    }

/*  for (curz = zstart; curz < M; curz++) */
    #pragma omp for schedule(dynamic,1)
    for (curz = zInnerLoopStart; curz < zInnerLoopUpper; curz++)
    {
      if (found || interrupted)
        continue;

#ifdef _OPENMP
     if (omp_get_thread_num() == 0)
#endif
     {
      if (utIsInterruptPending() != 0) {
	mexWarnMsgTxt("CTRL-C detected. Aborting...");
        flushMatlab();

        interrupted = true;
        continue;
      }

      if (display >= DISPLAY_VERBOSE)
      {
        #pragma omp critical
	{
          time2 = getcurtime();
          if (time2 - time1 > DISPLAY_INTERVAL)
          {
            time1 = time2;
            printf("%d %ld: ", d, M);
            for (tt = 0; tt < t; tt++)
              printf(" %ld ", z[tt]);
            printf(" ?%ld?\n", curz);
            flushMatlab();
          }
	}
      }
     }

      memset(bitmap, 0, bitmap_size);

      for (k = 0; k < l; k++)
      {
        int64_t ii = 0;
	uint64_t index;
        for (tt = 0; tt <= t-1; tt++)
	  ii += tempfreq[d*k+tt] * z[tt];
	ii += tempfreq[d*k+t] * curz;

        if (M_powerof_2)
	  index = modl2(ii, M_log2);
	else
          index = modl(ii, M);
       
        if (GET_BIT(bitmap, index) != 0)
	  break;
	SET_BIT(bitmap, index);
      }

      if (k == l)
      {
        #pragma omp critical
	{
	  if(display >= DISPLAY_VERBOSE)
    {
#ifdef _OPENMP
            printf("thread %d found z[%d]=%ld\n", omp_get_thread_num(), t, curz);
#else
            printf("found z[%d]=%ld\n", t, curz);
#endif
    }

	  if (found)
	  {
	    if (curz < z[t])
	      z[t] = curz;
	  }
	  else
	    z[t] = curz;

          found = true;
	}
      }
    }
    #pragma omp critical
    {
      mxFree(bitmap);
    }
 }
}

    if (interrupted)
    {
      mxFree(tempfreq);
      return -1;
    }

    if (!found)
    {
      mxFree(tempfreq);
      if (display >= DISPLAY_VERBOSE)
      {
        printf("Not found. Time: %.3e\n", getcurtime() - time0);
      }
      return 1;
    }

    if (display >= DISPLAY_CBC)
    {
      printf("z(%d) = %ld\n", t+1, z[t]);
      printf("z(1:%d) = [", t+1);
      for (tt = 0; tt <= t; tt++)
        printf(" %ld ", z[tt]);
      printf("]\n");
      flushMatlab();
    }
  }

  mxFree(tempfreq);

  if (display >= DISPLAY_CBC)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }
  return 0;
}

/**
  * Searches for non-aliasing generating vector z of rank-1 lattice for given
  * frequency index set and lattice size.
  */
static int search_gen_z_r1lattice2(int d, int64_t *z, int *freq, int len_freq, int64_t M, int inc, int nthreads, int display)
{
  int t, tt, k;
  int64_t curz;

  const uint64_t bitmap_length = 1+M/(1U<<6);
  const uint64_t bitmap_size = bitmap_length * sizeof(uint64_t);

  int64_t *tempsp = (int64_t *) mxMalloc(len_freq * sizeof(int64_t));
  double time0, time1, time2;
#ifdef GENLATTICE_FREQ_TRANSPOSE
  int *freqT = (int *) mxMalloc(d * len_freq * sizeof(int));
#endif
  
  int64_t M_log2 = 0;
  bool M_powerof_2 = false;

  volatile bool interrupted = false;

#ifdef GENLATTICE_FREQ_TRANSPOSE
  if (display >= DISPLAY_VERBOSE)
  {
    printf("transposed index set\n");
    flushMatlab();
  }
#endif

  if (display >= DISPLAY_VERBOSE)
  {
    printf("qsort start\n");
    flushMatlab();
  }

  time0 = getcurtime();

  compsize = d;
  qsort(freq, len_freq, d*sizeof(int), compfunc);

  if (display >= DISPLAY_VERBOSE)
  {
    printf("qsort finished\n");
    flushMatlab();
  }

#ifdef GENLATTICE_FREQ_TRANSPOSE
  for (k = 0; k < len_freq; k++)
    for (t = 0; t < d; t++)
      freqT[len_freq*t+k] = freq[d*k+t];
#endif

  for (k = 0; k < len_freq; k++)
#ifdef GENLATTICE_FREQ_TRANSPOSE
    tempsp[k] = freqT[k] * z[0];
#else
    tempsp[k] = freq[d*k] * z[0];
#endif

  for (t = 1; t < d; t++)
  {
    volatile bool found = false;
    uint64_t zstart = (inc == 1) ? max_i(z[t-1]+1, z[t]) : max_i(1, z[t]);
    uint64_t zInnerLoopStart;
    time1 = getcurtime();

    if (display >= DISPLAY_VERBOSE)
    {
      printf("%d %ld: ", d, M);
      for (tt = 0; tt < t; tt++)
        printf(" %ld ", z[tt]);
      printf(" ?%ld?\n", zstart);
      flushMatlab();
    }

for (zInnerLoopStart = zstart; zInnerLoopStart < M && !found && !interrupted; zInnerLoopStart += OMP_Z_INCR)
{
  uint64_t zInnerLoopUpper = min_i(M, zInnerLoopStart + OMP_Z_INCR);

 #pragma omp parallel shared(found,interrupted) private(tt,k) num_threads(nthreads)
 {
    uint64_t *bitmap;
    #pragma omp critical
    {
      bitmap = (uint64_t *) mxMalloc(bitmap_size);
    }

/** for (z[t] = zstart; z[t] < M; z[t]++)
  *   for (curz = zstart; curz < M; curz++)
  */
    #pragma omp for schedule(dynamic,1)
    for (curz = zInnerLoopStart; curz < zInnerLoopUpper; curz++)
    {
      uint64_t index;

      if (found || interrupted)
        continue;

#ifdef _OPENMP
     if (omp_get_thread_num() == 0)
#endif
     {
      if (utIsInterruptPending() != 0) {
	mexWarnMsgTxt("CTRL-C detected. Aborting...");
        flushMatlab();

        interrupted = true;
        continue;
      }

      if (display >= DISPLAY_VERBOSE)
      {
        #pragma omp critical
	{
          time2 = getcurtime();
          if (time2 - time1 > DISPLAY_INTERVAL)
          {
            time1 = time2;
            printf("%d %ld: ", d, M);
            for (tt = 0; tt < t; tt++)
              printf(" %ld ", z[tt]);
            printf(" ?%ld?\n", curz);
            flushMatlab();
          }
        }
      }
     }

      memset(bitmap, 0, bitmap_size);
#ifdef GENLATTICE_FREQ_TRANSPOSE
      index = modl(tempsp[0] + freqT[len_freq*t] * curz, M);
#else
      index = modl(tempsp[0] + freq[t] * curz, M);
#endif
      SET_BIT(bitmap, index);

      for (k = 1; k < len_freq; k++)
      {
#ifdef GENLATTICE_FREQ_TRANSPOSE
        if (tempsp[k-1] == tempsp[k] && freqT[len_freq*t+(k-1)] == freqT[len_freq*t+k])
	  continue;
#else
        if (tempsp[k-1] == tempsp[k] && freq[d*(k-1)+t] == freq[d*k+t])
	  continue;
#endif

#ifdef GENLATTICE_FREQ_TRANSPOSE
        if (M_powerof_2)
	  index = modl2(tempsp[k] + freqT[len_freq*t+k] * curz, M_log2);
	else
          index = modl(tempsp[k] + freqT[len_freq*t+k] * curz, M);
#else
        if (M_powerof_2)
	  index = modl2(tempsp[k] + freq[d*k+t] * curz, M_log2);
	else
          index = modl(tempsp[k] + freq[d*k+t] * curz, M);
#endif

        if (GET_BIT(bitmap, index) != 0)
	  break;
	SET_BIT(bitmap, index);
      }

      if (k == len_freq)
      {
        #pragma omp critical
	{
	  if(display >= DISPLAY_VERBOSE)
    {
#ifdef _OPENMP
            printf("thread %d found z[%d]=%ld\n", omp_get_thread_num(), t, curz);
#else
            printf("found z[%d]=%ld\n", t, curz);
#endif
    }

	  if (found)
	  {
	    if (curz < z[t])
	      z[t] = curz;
	  }
	  else
	    z[t] = curz;

          found = true;
	}
      }
    }
    #pragma omp critical
    {
      mxFree(bitmap);
    }
 }
}

    if (interrupted)
    {
      mxFree(tempsp);
#ifdef GENLATTICE_FREQ_TRANSPOSE
      mxFree(freqT);
#endif
      return -1;
    }

    if (!found)
    {
      mxFree(tempsp);
#ifdef GENLATTICE_FREQ_TRANSPOSE
      mxFree(freqT);
#endif
      if (display >= DISPLAY_VERBOSE)
      {
        printf("Not found. Time: %.3e\n", getcurtime() - time0);
      }
      return 1;
    }

    for (k = 0; k < len_freq; k++)
#ifdef GENLATTICE_FREQ_TRANSPOSE
      tempsp[k] += freqT[len_freq*t+k] * z[t];
#else
      tempsp[k] += freq[d*k+t] * z[t];
#endif

    if (display >= DISPLAY_CBC)
    {
      printf("z(%d) = %ld\n", t+1, z[t]);
      printf("z(1:%d) = [", t+1);
      for (tt = 0; tt <= t; tt++)
        printf(" %ld ", z[tt]);
      printf("]\n");
      flushMatlab();
    }
  }

  mxFree(tempsp);
#ifdef GENLATTICE_FREQ_TRANSPOSE
  mxFree(freqT);
#endif

  if (display >= DISPLAY_CBC)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }
  return 0;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  int nfreq;
  int d;
  int64_t M;
  double *outZ;
  int l, t, v;
  int *freq;
  int64_t *z;
  int inc = 1;
  int algorithmNr = GENLATTICE_DEFAULT_ALGORITHM;
  int display = DISPLAY_VERBOSE;
  int *cuboid_min, *cuboid_max;
  uint64_t cuboid_size;
#ifdef _OPENMP
  int nthreads = omp_get_max_threads();
#else
  int nthreads = 1;
#endif

  if(nrhs < 2)
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:nrhs",
                      "Two inputs required. Usage: z = gen_lattice_mex64_openmp(frequency_index_set, lattice_size_M)");

  if(nlhs!=1)
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:nlhs",
                      "One output required. Usage: z = gen_lattice_mex64_openmp(frequency_index_set, lattice_size_M)");

  if(mxGetNumberOfElements(prhs[1])!= 1 )
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:M_not_scalar",
                      "Input lattice size M (second argument) must be a scalar.");

  nfreq = mxGetM(prhs[0]);
  d = mxGetN(prhs[0]);
  if (mxGetScalar(prhs[1]) > (double) (1UL<<53))
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^53.");
    
  M = llrint(mxGetScalar(prhs[1]));

  if ((nrhs > 2 && mxIsStruct(prhs[2])) || (nrhs > 3 && mxIsStruct(prhs[3])))
  {
    int index = mxIsStruct(prhs[2]) ? 2 : 3;
    char str[256];
    mxArray *field;

    field = mxGetField(prhs[index], 0, "algorithm");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Algorithm");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "algorithm2") == 0 || strcmp(str, "Algorithm2") == 0 || strcmp(str, "2") == 0)
        algorithmNr = 2;
      else if (strcmp(str, "algorithm1") == 0 || strcmp(str, "Algorithm1") == 0 || strcmp(str, "1") == 0)
        algorithmNr = 1;
    }

    field = mxGetField(prhs[index], 0, "display");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Display");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "off") == 0 || strcmp(str, "Off") == 0
        ||strcmp(str, "none") == 0 || strcmp(str, "None") == 0)
        display = DISPLAY_OFF;
      else if (strcmp(str, "cbc") == 0 || strcmp(str, "Cbc") == 0 || strcmp(str, "CBC") == 0)
        display = DISPLAY_CBC;
    }

    field = mxGetField(prhs[index], 0, "inc");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Inc");
    if (field != NULL && lrint(mxGetScalar(field)) == 0)
      inc = 0;

    field = mxGetField(prhs[index], 0, "nthreads");
    if (field != NULL)
    {
#ifdef _OPENMP
      nthreads = min_i32(lrint(mxGetScalar(field)), omp_get_max_threads());
#else
      nthreads = min_i32(lrint(mxGetScalar(field)), 1);
#endif

      if (nthreads < 1)
        nthreads = 1;
    }
  }

  if (display > DISPLAY_OFF)
  {
    printf("d = %d, card_freq_index_set = %d, lattice_size_M = %ld, inc = %d, nthreads = %d, display = ", d, nfreq, M, inc, nthreads);
    if (display == DISPLAY_CBC)
      printf("cbc");
    else if (display == DISPLAY_VERBOSE)
      printf("verbose");
    printf("\n");
    flushMatlab();
  }

  if (M < nfreq)
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:M_less_than_card",
                      "Input lattice_size_M is smaller than cardinality of frequency index set.");

  z = (int64_t *) mxMalloc(d*sizeof(int64_t));
  z[0] = 1;
  for (t = 1; t < d; t++)
    z[t] = 0;

  if ((nrhs > 2 && !mxIsStruct(prhs[2])) || (nrhs > 3 && !mxIsStruct(prhs[3])))
  {
    int index = !mxIsStruct(prhs[2]) ? 2 : 3;
    int lzin = mxGetN(prhs[index]);
    v = convert_mxArrayCM_to_int64ArrayRM(z, prhs[index], 1, min_i(d,lzin));
    if (v != 0)
    {
      if (v == 1)
        mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:z_unsupported_type",
                          "Input z is of unsupported type. Either use double, single, int32, int16 or int8 array.");
      else if (v == 2)
        mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:z_not_real",
                          "Input z does not consist of real values.");
      else
        mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:z_unknown_error",
                          "Input z: unknown error.");
    }
  }

  if (display > DISPLAY_OFF)
  {
    printf("Initial z: ");
    for (t = 0; t < d; t++)
      printf("%ld ", z[t]);
    printf("\n");
  }

  freq = (int *) mxMalloc(d*nfreq*sizeof(int));

  v = convert_mxArrayCM_to_intArrayRM(freq, prhs[0], nfreq, d);
  if (v != 0)
  {
    if (v == 1)
      mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:freq_unsupported_type",
                        "Input frequency_index_set is of unsupported type. Either use double, single, int32, int16 or int8 array.");
    else if (v == 2)
      mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:freq_not_real",
                        "Input frequency_index_set does not consist of real values.");
    else
      mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:freq_unknown_error",
                        "Input frequency_index_set: unknown error.");

  }

  cuboid_min = (int *) mxMalloc(d*sizeof(int));
  cuboid_max = (int *) mxMalloc(d*sizeof(int));
  for (t = 0; t < d; t++)
  {
    cuboid_min[t] = freq[t];
    cuboid_max[t] = freq[t];
  }
  for (l = 1; l < nfreq; l++)
    for (t = 0; t < d; t++)
    {
      cuboid_min[t] = min_i(cuboid_min[t], freq[d*l+t]);
      cuboid_max[t] = max_i(cuboid_max[t], freq[d*l+t]);
    }

  cuboid_size = cuboid_max[0] - cuboid_min[0] + 1;
  for (t = 1; t < d; t++)
  {
    uint64_t factor = cuboid_max[t] - cuboid_min[t] + 1;
    if (factor * cuboid_size < cuboid_size)
    {
      cuboid_size = 0;
      break;
    }
    cuboid_size *= factor;
  }

  if (cuboid_size != 0 && M > cuboid_size)
  {
    if (display > DISPLAY_OFF)
    {
      printf("Cardinality of full cuboid-shaped grid in frequency domain: %ld, M=%ld\n", cuboid_size, M);
      mexWarnMsgTxt("Size of rank-1 lattice greater than full cuboid-shaped grid in frequency domain");
    }
  }

  if (algorithmNr == 2)
  {
    if (display > DISPLAY_OFF)
    {
      printf("Algorithm 2: Sort only once, operate on large index set array, incremental SP, ");
      printf("64-bit map");
      printf("\n");
      flushMatlab();
    }
    v = search_gen_z_r1lattice2(d, z, freq, nfreq, M, inc, nthreads, display);
  }
  else
  {
    if (display > DISPLAY_OFF)
    {
      printf("Algorithm 1: Re-sort index set for each additional component, non-incremental SP, ");
      printf("64-bit map");
      printf("\n");
      flushMatlab();
    }
    v = search_gen_z_r1lattice(d, z, freq, nfreq, M, inc, nthreads, display);
  }

  if (v == 0)
  {
    plhs[0] = mxCreateDoubleMatrix(1,d,mxREAL);
    outZ = mxGetPr(plhs[0]);
    for (t = 0; t < d; t++)
      outZ[t] = (double) z[t];

    if (display > DISPLAY_OFF)
    {
      printf("z = [", t+1);
      for (t = 0; t < d; t++)
        printf(" %ld ", z[t]);
      printf("]\n");
      flushMatlab();
    }
  }

  mxFree(cuboid_min);
  mxFree(cuboid_max);
  mxFree(freq);
  mxFree(z);

  if (v == -1)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex64_openmp:ctrl_c",
                      "CTRL-C pressed. Aborted.");
  }

  if (v != 0)
  {
    plhs[0] = mxCreateDoubleMatrix(0,0,mxREAL);
  }
}

