function [ freq, val, freq_cand_sizes, lattice_sizes, sample_sizes, z, run_times ] = sfft_cl_algorithm( fct_handle, d, N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, threshold_theta_basic, niter_r, niter_r_basic )
%SFFT_CL_ALGORITHM Sparse high-dimensional FFT based on rank-1 Chebyshev lattice sampling.
%   This is an approximate algorithm for the reconstruction of sparse
%   high-dimensional algebraic polynomials in Chebyshev basis, which adaptively
%   constructs the index set of frequencies belonging to the non-zero Chebyshev
%   coefficients in a dimension incremental way.
%   This method is based on the results in
%
%   D. Potts and T. Volkmer
%   Fast and exact reconstruction of arbitrary multivariate algebraic
%   polynomials in Chebyshev form
%   Proceedings of the 11th International Conference on Sampling Theory and
%   Applications, 392 - 396, 2015
%
%   and works analogously to the one presented in
%
%   D. Potts and T. Volkmer
%   Sparse high-dimensional FFT based on rank-1 lattice sampling
%   Submitted, 2015
%
%   Input parameters:
%      fct_handle - function handle (block box) which accepts sampling
%                   nodes from [-1,1]^d as an array of size [num_nodes,d]
%      d - dimensionality
%      N - parameter for search space Gamma = [0,...,N]^d of possible
%          frequencies
%      algorithm_name - (optional) type of algorithm,
%                       either 'a1cl' for Algorithm 1
%                       or 'a2cl' (default) for Algorithm 2
%      sparsity_s - (optional) target sparsity of output, default (N+1)^d
%      sparsity_s_local - (optional) value of sparsity_s in local loops,
%                         uses the value of sparsity_s as default
%      threshold_theta - (optional) relative threshold theta used for
%                        truncation in step 2e, default is 1e-12
%      threshold_theta_basic - (optional) relative threshold theta used for
%                              truncation in steps 1 and 2a,
%                              uses the value of threshold_theta as default 
%      niter_r - (optional) number of detection iteration r in steps 2b-2e,
%                default is 1
%      niter_r_basic - (optional) number of detection iteration r in
%                      steps 1 and 2a, uses the value of niter_r as default
%
%   Output parameters:
%      freq - detected frequencies belonging to the (approximately)
%             sparsity_s largest Chebyshev coefficients, size [sparsity_s,d]
%      val - corresponding Chebyshev coefficients
%      freq_cand_sizes - array of length d with number of investigated
%                        frequencies in each dimension increment step t
%      lattice_sizes - array of length d with size parameter M of
%                      reconstructing rank-1 Chebyshev lattices used for the
%                      sampling in each dimension increment step t
%      sample_sizes - array of length d with number of samples taken from
%                     the function fct_handle in each dimension increment
%                     step t
%      z - generating vector z of reconstructing rank-1 Chebyshev lattice for
%          frequencies in freq, lattice_sizes(end) contains the
%          corresponding size parameter M of the rank-1 Chebyshev lattice
%      run_times - array of length 5 with runtimes measured for various
%                  parts of the algorithm,
%                  run_times(1): steps 1 and 2a,
%                  run_times(2): building of reconstructing rank-1 Chebyshev
%                                lattices
%                  run_times(3): sampling step 2c
%                  run_times(4): inverse rank-1 Chebyshev lattice FFT step 2d
%                  run_times(5): total runtime of the algorithm
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

warning('off', 'MATLAB:maxNumCompThreads:Deprecated');
nthreads = min(12,maxNumCompThreads);

if nargin < 4
  algorithm_name = 'a2cl';
end

if nargin < 5
  sparsity_s = (N+1)^d;
end

if nargin < 6
  sparsity_s_local = sparsity_s;
end

if nargin < 7
  threshold_theta = 1e-12;
end

if nargin < 8
  threshold_theta_basic = threshold_theta;
end

if nargin < 9
  niter_r = 1;
end

if nargin < 10
  niter_r_basic = niter_r;
end

check_naturalnumber_or_error(d, 'dimensionality d');
check_naturalnumber_or_error(N, 'parameter N');

Gamma_symhc = 0;
if strcmp(algorithm_name, 'a1clsymhc')
  algorithm_name = 'a1cl';
  Gamma_symhc = 1;
end

switch algorithm_name
  case {'a1cl','a2cl','a1clr1l'}
  otherwise
    error('unsupported algorithm: %s', algorithm_name);
end

check_naturalnumber_or_error(sparsity_s, 'parameter sparsity_s');
check_naturalnumber_or_error(sparsity_s_local, 'parameter sparsity_s_local');
check_0leq_param_leq1_or_error(threshold_theta, 'parameter threshold_theta');
check_0leq_param_leq1_or_error(threshold_theta_basic, 'parameter threshold_theta_basic');
check_naturalnumber_or_error(niter_r, 'parameter niter_r');
check_naturalnumber_or_error(niter_r_basic, 'parameter niter_r_basic');

check_mex_file_or_warn('dct_cbc_lattice_search_mex64_openmp');
check_mex_file_or_warn('dct_lattice_size_reduce_mex64_openmp');
check_mex_file_or_warn('dct_lattice_size_reduce_bisect_only_mex');
check_mex_file_or_warn('lfft_gen_lattice_mex');
check_mex_file_or_warn('lfft_lattice_size_reduce_even_mex');
check_mex_file_or_warn('lfft_lattice_size_reduce_bisect_only_mex');
check_mex_file_or_warn('lfft_lattice_size_reduce_mex');

fprintf('sfft_cl: d = %d, N = %d, algorithm = %s, sparsity_s = %d, sparsity_s_local = %d\n', d, N, algorithm_name, sparsity_s, sparsity_s_local);
fprintf('sfft_cl: threshold_theta = %.3e, basic_threshold_theta = %.3e, niter_r = %d, niter_r_basic = %d\n', threshold_theta, threshold_theta_basic, niter_r, niter_r_basic);

freq_cand_sizes = zeros(1,d);
lattice_sizes = zeros(1,d);
sample_sizes = zeros(1,d);
run_times = zeros(1,5); % basic, lattice search, sampling, lfft, all

freq_cand_sizes(1) = N+1;
lattice_sizes(1) = N+1;

tAll = tic;

if d == 1
  tBasic = tic;
  [ freq, val ] = reco_cardI_cl_basic1( fct_handle, d, N, 1, threshold_theta_basic);
  run_times(1) = toc(tBasic);
  sample_sizes(1) = N+2;
else
  freq = zeros(1,1);
  z = zeros(1,d);
  
  % step 1
  L_1 = N+1;
  for vv=1:niter_r_basic
    tBasic = tic;
    [ freq_local, ~ ] = reco_cardI_cl_basic1( fct_handle, d, N, 1, threshold_theta_basic);
    run_times(1) = run_times(1) + toc(tBasic);
    sample_sizes(1) = sample_sizes(1) + N+2;
    freq = union(freq,freq_local,'rows');
  end
  
  freq_tilde = indexset_mirror_all(freq);

  if strcmp(algorithm_name, 'a1cl') || strcmp(algorithm_name, 'a2cl')
    for S_1=1:L_1
      if length(unique(mod(freq_tilde,2*S_1))) == length(freq_tilde)
        break;
      end
    end
  elseif strcmp(algorithm_name, 'a1clr1l')
    for S_1=1:(2*L_1+1)
      if length(unique(mod(freq_tilde,S_1))) == length(freq_tilde)
        break;
      end
    end
  else
    error('unknown algo');
  end
  
  clear freq_tilde

  fprintf('step 1: L_1 = %d, card(I^(1)) = %d, S_1 = %d\n', L_1, size(freq,1), S_1);
  
  if S_1 == 1
    z(1) = 0;
  else
    z(1) = 1;
  end
  M = S_1;
  M_old = M;
  z_old = z;
  M_tilde = 2*S_1;
  % step 2
  for t = 2:d
    fprintf('step 2: t = %d start\n', t);
    freq_t = zeros(1,1);
    % step 2a
    L_t = N+1;
    for vv=1:niter_r_basic
      tBasic = tic;
      [ freqnew_local, ~ ] = reco_cardI_cl_basic1( fct_handle, d, N, t, threshold_theta_basic);
      run_times(1) = run_times(1) + toc(tBasic);
      sample_sizes(t) = sample_sizes(t) + N+2;
      freq_t = union(freq_t,freqnew_local,'rows');
    end
    fprintf('step 2a: L_%d = %d, card(I^(%d)) = %d\n', t, L_t, t, size(freq_t,1));
    A=reshape(repmat(1:size(freq,1), size(freq_t,1), 1), size(freq,1)*size(freq_t,1), 1);
    B=repmat((1:size(freq_t,1))', size(freq,1), 1);
    indcross = [A B];
    clear A B;
    freq_cand = [freq(indcross(:,1),:) freq_t(indcross(:,2),:)];
    clear indcross;
    
    if Gamma_symhc
      freq_cand = freq_cand(prod(max(1,freq_cand),2) <= N,:);
    end

    freq_cand_sizes(t) = size(freq_cand,1);

    % step 2b
    if t==d
      niter_r_tilde=1;
    else
      niter_r_tilde=niter_r;
    end

    freq_t_tilde = indexset_mirror_all(freq_t);
    
    if strcmp(algorithm_name, 'a1cl') || strcmp(algorithm_name, 'a2cl')
      for S_t=1:L_t
        if length(unique(mod(freq_t_tilde,2*S_t))) == length(freq_t_tilde)
          break;
        end
      end
    elseif strcmp(algorithm_name, 'a1clr1l')
      for S_t=1:(2*L_t+1)
        if length(unique(mod(freq_t_tilde,S_t))) == length(freq_t_tilde)
          break;
        end
      end
    else
      error('unknown algo');
    end
    clear freq_t_tilde

    fprintf('step 2b: r_tilde = %d, S_%d = %d\n', niter_r_tilde, t, S_t);

    if strcmp(algorithm_name, 'a1cl')
      fprintf('step 2b: start building reconstructing rank-1 lattice\n');
      tLattice = tic;
      optsSearch.Display='cbc';
      optsSearch.Inc=0;
      if nthreads > 0
        optsSearch.nthreads = nthreads;
      end
      %Mcube = prod(max(freq_cand,[],1)-min(freq_cand,[],1)+1);
%      Mcube = prod( 2*ceil((max(abs(freq_cand),[],1)+min(abs(freq_cand),[],1)+1)/2) );
      Min = 2*M * 2*S_t; %min(2*M * S_t, 2*Mcube);
      z_out = [];
      fprintf('checking for easy z\n');
      if dct_r1l_check(freq_cand,[z(1:t-1), 2*M],lattice_sizes(t-1),1)
        z_out = [z(1:t-1), 2*M];
        fprintf('success1 easy z\n');
        Min = lattice_sizes(t-1);
      elseif dct_r1l_check(freq_cand,[z(1:t-1), 2*M],Min,1)
        z_out = [z(1:t-1), 2*M];
        fprintf('success2 easy z\n');
      end
      while length(z_out) ~= t
        if exist(strcat('dct_cbc_lattice_search_mex64_openmp.',mexext), 'file') == 3
          z_out = dct_cbc_lattice_search_mex64_openmp(freq_cand,Min,optsSearch,z(1:t-1));
        else
          z_out = dct_cbc_lattice_search(freq_cand,Min); %,optsSearch,z(1:t-1));
        end
        if length(z_out) ~= t
          fprintf('no generating vector found, increasing M from %d to %d\n', Min, simplenextprime(2*Min));
          Min = simplenextprime(2*Min);
        end
      end
      z(1:t) = z_out;
      if ~dct_r1l_check(freq_cand,z(1:t),Min)
        error('Invalid generating vector');
      end

if Min == lattice_sizes(t-1)
  P_out = Min;
else
      optPmin.Display='iter';
      optPmin.Prime=0;
      optPmin.Bisection_Successors=100;

      if nthreads > 0
        optPmin.nthreads = nthreads;
      end

      P_in=Min;
      if exist(strcat('dct_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
        P_out=dct_lattice_size_reduce_bisect_only_mex(freq_cand,P_in,z(1:t),optPmin);
      else
        P_out = dct_lattice_size_reduce_bisect_only(freq_cand,P_in,z(1:t),optPmin);
      end
      fprintf('step 2b:   np start, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
      P_in=P_out;

      optPmin.Display='iter';
      optPmin.Prime=1;
      optPmin.Bisection_Successors=100;
      for r=1:10
        if exist(strcat('dct_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
          P_out=dct_lattice_size_reduce_bisect_only_mex(freq_cand,P_in,z(1:t),optPmin);
        else
          P_out = dct_lattice_size_reduce_bisect_only(freq_cand,P_in,z(1:t),optPmin);
        end
        fprintf('step 2b:    p r =%2d, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', r, P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
        if P_in==P_out
          break;
        end
        P_in = P_out;
      end
end
      M=P_out;
      run_times(2) = run_times(2) + toc(tLattice);
      fprintf('step 2b: finished building reconstructing rank-1 lattice\n');
    elseif strcmp(algorithm_name, 'a2cl')
      z(t) = M_tilde;
      M_tilde = M_tilde * 2*S_t;
      M = M_tilde / 2;
    elseif strcmp(algorithm_name, 'a1clr1l')
      fprintf('step 2b: start building reconstructing rank-1 lattice\n');
      tLattice = tic;
      optsSearch.Display='off';
      optsSearch.Inc=0;
      if nthreads > 0
        optsSearch.nthreads = nthreads;
      end
      freq_cand_tilde = indexset_mirror_all(freq_cand);
      Mcube = prod(max(freq_cand_tilde,[],1)-min(freq_cand_tilde,[],1)+1);
      Min = min(2 * M * S_t, Mcube);
      if exist(strcat('lfft_gen_lattice_mex.',mexext), 'file') == 3
        try
          z_out = lfft_gen_lattice_mex(freq_cand_tilde,Min,optsSearch,z(1:t-1));
        catch
          z_out = [];
          warning('lfft_gen_lattice_mex failed');
        end
        if length(z_out) ~= t
          warning('z_out too short, using z(t) = M, newM = M * S_t');
          z_out = [z(1:t-1), M];
          Min = M * S_t;
        end
      else
        try
          z_out = lfft_gen_lattice(freq_cand_tilde,Min,optsSearch,z(1:t-1));
        catch
          z_out = [];
          warning('lfft_gen_lattice failed');
        end
        if length(z_out) ~= t
          warning('z_out too short, using z(t) = M, newM = M * S_t');
          z_out = [z(1:t-1), M];
          Min = M * S_t;
        end
      end
      if length(z_out) ~= t
        error('length of z');
      end
      z(1:t) = z_out;
      if length(unique(mod(freq_cand_tilde*z(1:t)',Min))) ~= size(freq_cand_tilde,1)
        error('Invalid generating vector');
      end

      optPmin.Display='off';
      optPmin.Prime=0;
      optPmin.Bisection_Successors=1000;

      if nthreads > 0
        optPmin.nthreads = nthreads;
      end

      P_in=Min;
      if exist(strcat('lfft_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
        P_out=lfft_lattice_size_reduce_bisect_only_mex(freq_cand_tilde,P_in,z(1:t),optPmin);
      else
        P_out = lfft_lattice_size_reduce_bisect_only(freq_cand_tilde,P_in,z(1:t),optPmin);
      end
      fprintf('step 2b:   np start, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
      P_in=P_out;

      optPmin.Display='off';
      optPmin.Prime=1;
      optPmin.Bisection_Successors=1000;
      for r=1:10
        if exist(strcat('lfft_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
          P_out=lfft_lattice_size_reduce_bisect_only_mex(freq_cand_tilde,P_in,z(1:t),optPmin);
        else
          P_out = lfft_lattice_size_reduce_bisect_only(freq_cand_tilde,P_in,z(1:t),optPmin);
        end
        fprintf('step 2b:    p r =%2d, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', r, P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
        if P_in==P_out
          break;
        end
        P_in = P_out;
      end
      M=P_out;
      
%       while mod(M,2) ~= 0
%         optsReduce.nthreads = nthreads;
%         optsReduce.Mmin = M+1;
%         optsReduce.increase = 1;
%         optsReduce.Display='off';
%         optsReduce.prime = 0;
%         if exist(strcat('lfft_lattice_size_reduce_mex.',mexext), 'file') == 3
%           M = lfft_lattice_size_reduce_mex(freq_cand_tilde, 2*M, z(1:t), optsReduce);
%         else
%           M = lfft_lattice_size_reduce(freq_cand_tilde, 2*M, z(1:t), optsReduce);
%         end
%       end
        if mod(M,2) ~= 0
          M = 2 * M;
        end

        if length(unique(mod(freq_cand_tilde*z(1:t)',M))) ~= size(freq_cand_tilde,1)
          error('rank-1 lattice does not fulfill reconstructing property');
        end
        
        clear freq_cand_tilde
        M = M/2;
        if ~dct_r1l_check(freq_cand,z(1:t),M)
          error('rank-1 Chebyshev lattice does not fulfill reconstructing property');
        end
      
      run_times(2) = run_times(2) + toc(tLattice);
      fprintf('step 2b: finished building reconstructing rank-1 lattice\n');
    else
      error('not supported');
    end

    fprintf('step 2b: z_%d = %d, z(1:%d) = [', t, z(t), t);
    for s=1:t
      fprintf(' %d', z(s));
    end
    fprintf(' ], M = %d\n', M);

    if ~dct_r1l_check(freq_cand,z(1:t),M)
      error('rank-1 Chebyshev lattice does not fulfill reconstructing property');
    end

    lattice_sizes(t) = M;

    fprintf('step 2c: %d samples, r_tilde = %d detection iterations\n', M+1, niter_r_tilde);
    %drawnow;

    ind = [];
    for vv=1:niter_r_tilde
      % step 2c
      tSampling = tic;
      try
        xx = 2*rand(d-t,1)-1;
        samples = fct_handle(z(1:t),M,xx);
% 
%         xj = zeros(M+1,d);
%         xj(:,t+1:d) = repmat(xx',M+1,1);
%         xj(:,1:t) = cos((0:M)' * z(1:t) * pi / M);
%         samples2 = fct_handle(xj);
%         clear xj;
%         max(abs(samples-samples2));
      catch
%        warning('using fallback: slow sampling');
        xj = repmat(2*rand(1,d)-1,M+1,1);
        xj(:,1:t) = cos((0:M)' * z(1:t) * pi / M);
        samples = fct_handle(xj);
        clear xj;
      end
      run_times(3) = run_times(3) + toc(tSampling);
      sample_sizes(t) = sample_sizes(t) + M+1;

      % step 2d
      tLFFT = tic;
      val_all = cheb_poly_reco(freq_cand,z(1:t),M,samples);
      run_times(4) = run_times(4) + toc(tLFFT);
      clear samples;

      % step 2e
      max_abs_val_all = max(abs(val_all));
      absolute_threshold = threshold_theta*max_abs_val_all;
      ind_local = find(abs(val_all) >= absolute_threshold);
      val_local = val_all(abs(val_all) >= absolute_threshold);
      clear max_abs_val_all;
      if length(ind_local) > sparsity_s_local
        [~,ind_local] = sort(abs(val_local),'descend');
        ind_local = ind_local(1:sparsity_s_local);
      end
      fprintf('step 2e: i = %d, max fc_all = %.1e, min fc_all = %.1e, max fc_local = %.1e, min fc_local = %.1e\n', vv, max(abs(val_all)), min(abs(val_all)), max(abs(val_all(ind_local))), min(abs(val_all(ind_local))));
      clear val_local;
      if vv==1
        ind = ind_local;
      else
        ind = union(ind,ind_local,'rows');
      end
      clear ind_local;

      fprintf('step 2e: i = %d, card(I^(1,...,%d)) = %d\n', vv, t, length(ind));
    end
    freq = freq_cand(ind,:);
    val = val_all(ind);
    if ((t==d) && (length(val) > sparsity_s))
      [~,ind] = sort(abs(val),'descend');
      val = val(ind(1:sparsity_s));
      freq = freq(ind(1:sparsity_s),:);
    end
    fprintf('step 2e: result dim %d of %d: %d freq candidates, %d samples, %d freq detected\n', t, d, size(freq_cand,1), M+1, size(freq,1));
    clear ind;

    % additional step 2f
    if t ~= d
      if strcmp(algorithm_name, 'a1cl')
if (M_old < M) && dct_r1l_check(freq,z_old(1:t),M_old)
          z = z_old;
          M = M_old;
        fprintf('additional step 2f: z_%d = %d, z(1:%d) = [', t, z(t), t);
        for s=1:t
          fprintf(' %d', z(s));
        end
        fprintf(' ], M_%d = %d\n', t, M);
else
        tLattice = tic;
        %Mcube = prod( 2*ceil((max(abs(freq),[],1)+min(abs(freq),[],1)+1)/2) );
        Min = M; %min(M, 2*Mcube);
        optsSearch.Display='off';
        optsSearch.Inc=0;
        if nthreads > 0
          optsSearch.nthreads = nthreads;
        end
        if exist(strcat('dct_cbc_lattice_search_mex64_openmp.',mexext), 'file') == 3
          z_out = dct_cbc_lattice_search_mex64_openmp(freq,Min,optsSearch,z(1:t-1));
        else
          z_out = dct_cbc_lattice_search(freq,Min); %,optsSearch,z(1:t-1));
        end
        while length(z_out) ~= t
          Min = simplenextprime(2*Min);
          if exist(strcat('dct_cbc_lattice_search_mex64_openmp.',mexext), 'file') == 3
            z_out = dct_cbc_lattice_search_mex64_openmp(freq,Min,optsSearch,z(1:t-1));
          else
            z_out = dct_cbc_lattice_search(freq,Min); %,optsSearch,z(1:t-1));
          end
        end
        z(1:t) = z_out;
        if ~dct_r1l_check(freq,z(1:t),Min)
          error('Invalid generating vector');
        end

        optsReduce.Display='off';
        if t==2
          optsReduce.Mmin = size(freq,1);
        else
          optsReduce.Mmin = min(size(freq,1),lattice_sizes(t-1));
        end
        if nthreads > 0
          optsReduce.nthreads = nthreads;
        end
        if exist(strcat('dct_lattice_size_reduce_mex64_openmp.',mexext), 'file') == 3
          M = dct_lattice_size_reduce_mex64_openmp(freq,Min,z(1:t),optsReduce);
        else
          M = dct_lattice_size_reduce(freq,Min,z(1:t));
        end
        run_times(2) = run_times(2) + toc(tLattice);

        fprintf('additional step 2f: z_%d = %d, z(1:%d) = [', t, z(t), t);
        for s=1:t
          fprintf(' %d', z(s));
        end
        fprintf(' ], M_%d = %d\n', t, M);

        if ~dct_r1l_check(freq,z(1:t),M)
          error('rank-1 Chebyshev lattice does not fulfill reconstructing property');
        end
end
      elseif strcmp(algorithm_name, 'a2cl')
        tLattice = tic;
        freq_tilde = indexset_mirror_all(freq);
        %Mcube_tilde = prod( 2*ceil((max(freq_tilde,[],1)-min(freq_tilde,[],1)+1)/2) );
        Min_tilde = M_tilde; %min(M_tilde, Mcube_tilde);
        optsSearch.Display='off';
        optsSearch.Inc=0;
        if nthreads > 0
          optsSearch.nthreads = nthreads;
        end
        if exist(strcat('lfft_gen_lattice_mex.',mexext), 'file') == 3
          z_out = lfft_gen_lattice_mex(freq_tilde,Min_tilde,optsSearch,z(1:t-1));
        else
          z_out = lfft_gen_lattice(freq_tilde,Min_tilde,optsSearch,z(1:t-1));
        end
        if length(z_out) ~= t
          error('length of z');
        end
        z(1:t) = z_out;
        if length(unique(mod(freq_tilde*z(1:t)',Min_tilde))) ~= size(freq_tilde,1)
          error('Invalid generating vector');
        end

        optsReduce.Display='off';
        if t==2
          optsReduce.Mmin = size(freq_tilde,1);
        else
          optsReduce.Mmin = min(size(freq_tilde,1),lattice_sizes(t-1));
        end
        if nthreads > 0
          optsReduce.nthreads = nthreads;
        end
        if exist(strcat('lfft_lattice_size_reduce_even_mex.',mexext), 'file') == 3
          M_tilde = lfft_lattice_size_reduce_even_mex(freq_tilde,Min_tilde,z(1:t),optsReduce);
        else
          M_tilde = lfft_lattice_size_reduce_even(freq_tilde,Min_tilde,z(1:t));
        end
        run_times(2) = run_times(2) + toc(tLattice);

        fprintf('additional step 2f: z_%d = %d, z(1:%d) = [', t, z(t), t);
        for s=1:t
          fprintf(' %d', z(s));
        end
        fprintf(' ], M_tilde_%d = %d\n', t, M_tilde);

        if mod(M_tilde,2) ~= 0
          error('internal error: M_tilde should have been an even natural number');
        end
        
        if length(unique(mod(freq_tilde*z(1:t)',M_tilde))) ~= size(freq_tilde,1)
          error('rank-1 lattice does not fulfill reconstructing property');
        end
        
        M = M_tilde / 2;
        fprintf('M_%d = %d\n', t, M);
        clear freq_tilde
      elseif strcmp(algorithm_name, 'a1clr1l')
        M = 2 * M;
        freq_tilde = indexset_mirror_all(freq);
%         if (M_old < M) && (length(unique(mod(freq_tilde*z_old(1:t)',M_old))) == size(freq_tilde,1))
%           z = z_old;
%           M = M_old;
%         end
        freq_lengths = max(freq_tilde,[],1)-min(freq_tilde,[],1)+1;
        Mcube = prod(freq_lengths);
        Min = min(M, Mcube);
        optsSearch.Display='off';
        optsSearch.Inc=0;
        if nthreads > 0
          optsSearch.nthreads = nthreads;
        end
        if Min < M
          z_out = ones(1,t);
          for j = 2:t
            z_out(j) = z_out(j-1) * freq_lengths(j-1);
          end
        else
          if exist(strcat('lfft_gen_lattice_mex.',mexext), 'file') == 3
            z_out = lfft_gen_lattice_mex(freq_tilde,Min,optsSearch,z(1:t-1));
          else
            z_out = lfft_gen_lattice(freq_tilde,Min,optsSearch,z(1:t-1));
          end
        end
        if length(z_out) ~= t
          error('length of z');
        end
        z(1:t) = z_out;
        if length(unique(mod(freq_tilde*z(1:t)',Min))) ~= size(freq_tilde,1)
          error('Invalid generating vector');
        end

        optsReduce.Display='off';
        optsReduce.Prime=1;
        optsReduce.Bisection_Successors=1000;
        if t==2
          optsReduce.Mmin = size(freq_tilde,1);
        else
          optsReduce.Mmin = min(size(freq_tilde,1),lattice_sizes(t-1));
        end
        if nthreads > 0
          optsReduce.nthreads = nthreads;
        end
        if exist(strcat('lfft_lattice_size_reduce_mex.',mexext), 'file') == 3
          M = lfft_lattice_size_reduce_mex(freq_tilde,Min,z(1:t),optsReduce);
        else
          M = lfft_lattice_size_reduce(freq_tilde,Min,z(1:t),optsReduce);
        end
        run_times(2) = run_times(2) + toc(tLattice);

        fprintf('additional step 2f: z_%d = %d, z(1:%d) = [', t, z(t), t);
        for s=1:t
          fprintf(' %d', z(s));
        end
        fprintf(' ], M_%d = %d\n', t, M);

%         while mod(M,2) ~= 0
%           optsReduce.nthreads = nthreads;
%           optsReduce.Mmin = M+1;
%           optsReduce.increase = 1;
%           optsReduce.Display='off';
%           optsReduce.prime = 0;
%           if exist(strcat('lfft_lattice_size_reduce_mex.',mexext), 'file') == 3
%             M = lfft_lattice_size_reduce_mex(freq_tilde, 2*M, z(1:t), optsReduce);
%           else
%             M = lfft_lattice_size_reduce(freq_tilde, 2*M, z(1:t), optsReduce);
%           end
%         end
        if mod(M,2) ~= 0
          M = 2 * M;
        end

        if length(unique(mod(freq_tilde*z(1:t)',M))) ~= size(freq_tilde,1)
          error('rank-1 lattice does not fulfill reconstructing property');
        end
        clear freq_tilde
        M = M / 2;
        if ~dct_r1l_check(freq,z(1:t),M)
          error('rank-1 Chebyshev lattice does not fulfill reconstructing property');
        end
      else
        error('not supported');
      end
    end
    fprintf('step 2: t = %d end, d = %d, M = %d\n\n', t, d, M);
    z_old = z;
    M_old = M;
  end
  fprintf('step 3: %.10g possible freq., %d total samples, %d freq. detected\n', (N+1)^d, sum(sample_sizes), size(freq,1));
end

run_times(end) = toc(tAll);

end

function [val] = check_naturalnumber_or_error(param, name)
  val = 0;
  if (param < 1) || (param ~= int32(param))
    error('%s must be a natural number', name);
  end
end

function [val] = check_0leq_param_leq1_or_error(param, name)
  val = 0;
  if (param < 0) || (param > 1)
    error('%s must be >= 0 and <= 1', name);
  end
end

function [filename] = check_mex_file_or_warn(filename_without_ext)
  filename = strcat(filename_without_ext,strcat('.',mexext));
  if exist(filename, 'file') ~= 3
    warning('MATLAB mex file %s is missing. The algorithm will be very slow!\nPlease use the MATLAB script compile_mex_files_without_openmp.m or compile_mex_files_with_openmp.m', filename);
  end
end

function [y] = simplenextprime(n)
% computes the next prime larger n
% uses simple isprime function of MATLAB
% Author: Lutz Kämmerer

y=n+1;
if mod(y,2)==0,
   y=y+1;
end;

flag=1;
while flag,
   if(isprime(y)),
      return;
   end;
   y=y+2;
end;
end
